from __future__ import annotations

import os
from typing import Any, Dict, List, Optional, Text, Union

import structlog

from rasa.shared.utils.io import resolve_environment_variables
from rasa.shared.utils.yaml import read_config_file
from rasa.tracing.constants import (
    ENDPOINTS_TRACING_KEY,
    LANGFUSE_CONFIG_BASE_URL_KEY,
    LANGFUSE_CONFIG_DEBUG_KEY,
    LANGFUSE_CONFIG_ENVIRONMENT_KEY,
    LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY,
    LANGFUSE_CONFIG_PRIVATE_KEY_KEY,
    LANGFUSE_CONFIG_PUBLIC_KEY_KEY,
    LANGFUSE_CONFIG_RELEASE_KEY,
    LANGFUSE_CONFIG_SAMPLE_RATE_KEY,
    LANGFUSE_CONFIG_TIMEOUT_KEY,
    TRACING_TYPE_LANGFUSE,
)
from rasa.tracing.exceptions import (
    DuplicateTracingConfigException,
    InvalidLangfuseConfigException,
)
from rasa.utils.endpoints import EndpointConfig

structlogger = structlog.get_logger()


def configure_langfuse(endpoints_file: str) -> None:
    """Configure Langfuse callback for litellm from endpoints file."""
    langfuse_config = _get_langfuse_config(endpoints_file)

    if not langfuse_config:
        structlogger.debug(
            "langfuse_configuration.langfuse_config_not_found",
            event_info=(
                f"No Langfuse configuration found in {endpoints_file}, "
                f"Langfuse will not be configured."
            ),
        )
        return

    if not _is_langfuse_available():
        return

    config_values = _extract_langfuse_config_values(langfuse_config)

    _validate_langfuse_config(config_values)

    public_key = config_values[LANGFUSE_CONFIG_PUBLIC_KEY_KEY]
    private_key = config_values[LANGFUSE_CONFIG_PRIVATE_KEY_KEY]

    resolved_keys = _resolve_environment_variables(public_key, private_key)
    _set_langfuse_environment_variables(config_values, resolved_keys)
    _configure_litellm_callback()

    structlogger.info(
        "langfuse_configuration.langfuse_config_configured",
        event_info=(f"Langfuse has been configured from {endpoints_file}."),
    )


def _is_langfuse_available() -> bool:
    """Check if Langfuse is available."""
    try:
        import langfuse  # noqa: F401
    except ImportError as e:
        structlogger.error(
            "langfuse_configuration.langfuse_not_available",
            event_info=(
                "Langfuse is not available. "
                "Please install langfuse via `pip install rasa-pro[monitoring]`. "
                "Langfuse will not be configured."
            ),
            error=str(e),
        )
        return False
    return True


def _get_langfuse_config(endpoints_file: str) -> Optional[EndpointConfig]:
    """Read Langfuse configuration from endpoints file."""
    if not endpoints_file:
        return None

    try:
        content = read_config_file(endpoints_file)
    except FileNotFoundError:
        structlogger.debug(
            "langfuse_configuration.endpoints_file_not_found",
            event_info=(
                f"Endpoints file not found: {endpoints_file}, "
                f"Langfuse will not be configured."
            ),
        )
        return None
    except Exception:
        structlogger.error(
            "langfuse_configuration.endpoints_file_read_error",
            event_info=(
                f"Error reading endpoints file: {endpoints_file}, "
                f"Langfuse will not be configured."
            ),
        )
        return None

    if content.get(ENDPOINTS_TRACING_KEY) is None:
        return None

    tracing_config: Union[List[Dict[Text, Any]], Dict[Text, Any]] = content.get(
        ENDPOINTS_TRACING_KEY, []
    )

    tracing_configs = _parse_tracing_configs(tracing_config)

    langfuse_configs = [
        config for config in tracing_configs if config.type == TRACING_TYPE_LANGFUSE
    ]

    if len(langfuse_configs) == 0:
        return None

    if len(langfuse_configs) > 1:
        _log_multiple_langfuse_configs_error(endpoints_file)
        raise DuplicateTracingConfigException(
            f"Multiple Langfuse configs found in {endpoints_file}, "
            "which is not supported. Only one Langfuse config is allowed."
        )

    return langfuse_configs[0]


def _parse_tracing_configs(
    tracing_config: Union[List[Dict[Text, Any]], Dict[Text, Any]],
) -> List[EndpointConfig]:
    """Parse tracing configuration into list of EndpointConfig objects."""
    if isinstance(tracing_config, list):
        return [EndpointConfig.from_dict(item) for item in tracing_config]
    else:
        return [EndpointConfig.from_dict(tracing_config)]


def _log_multiple_langfuse_configs_error(endpoints_file: str) -> None:
    """Log error when multiple Langfuse configs are found."""
    structlogger.error(
        "langfuse_configuration.multiple_langfuse_configs",
        event_info=(
            f"Multiple Langfuse configs found in {endpoints_file}, "
            f"which is not supported. Only one Langfuse config is allowed."
        ),
    )


def _extract_langfuse_config_values(
    langfuse_config: EndpointConfig,
) -> Dict[str, Any]:
    """Extract all configuration values from Langfuse config."""
    return {
        LANGFUSE_CONFIG_PUBLIC_KEY_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_PUBLIC_KEY_KEY
        ),
        LANGFUSE_CONFIG_PRIVATE_KEY_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_PRIVATE_KEY_KEY
        ),
        LANGFUSE_CONFIG_BASE_URL_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_BASE_URL_KEY
        ),
        LANGFUSE_CONFIG_TIMEOUT_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_TIMEOUT_KEY
        ),
        LANGFUSE_CONFIG_DEBUG_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_DEBUG_KEY
        ),
        LANGFUSE_CONFIG_ENVIRONMENT_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_ENVIRONMENT_KEY
        ),
        LANGFUSE_CONFIG_RELEASE_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_RELEASE_KEY
        ),
        LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY
        ),
        LANGFUSE_CONFIG_SAMPLE_RATE_KEY: langfuse_config.kwargs.get(
            LANGFUSE_CONFIG_SAMPLE_RATE_KEY
        ),
    }


def _resolve_environment_variables(
    public_key: str, private_key: str
) -> Dict[str, Union[str, List[Any], Dict[str, Any]]]:
    """Resolve environment variables in public and secret keys."""
    return {
        LANGFUSE_CONFIG_PUBLIC_KEY_KEY: resolve_environment_variables(public_key),
        LANGFUSE_CONFIG_PRIVATE_KEY_KEY: resolve_environment_variables(private_key),
    }


def _set_langfuse_environment_variables(
    config_values: Dict[str, Any], resolved_keys: Dict[str, Optional[str]]
) -> None:
    """Set Langfuse environment variables for LiteLLM integration."""
    if resolved_keys[LANGFUSE_CONFIG_PUBLIC_KEY_KEY] is not None:
        os.environ["LANGFUSE_PUBLIC_KEY"] = resolved_keys[
            LANGFUSE_CONFIG_PUBLIC_KEY_KEY
        ]  # type: ignore[assignment]
    if resolved_keys[LANGFUSE_CONFIG_PRIVATE_KEY_KEY] is not None:
        os.environ["LANGFUSE_SECRET_KEY"] = resolved_keys[
            LANGFUSE_CONFIG_PRIVATE_KEY_KEY
        ]  # type: ignore[assignment]
    if config_values[LANGFUSE_CONFIG_BASE_URL_KEY] is not None:
        os.environ["LANGFUSE_OTEL_HOST"] = config_values[LANGFUSE_CONFIG_BASE_URL_KEY]
    if config_values[LANGFUSE_CONFIG_TIMEOUT_KEY] is not None:
        os.environ["LANGFUSE_TIMEOUT"] = config_values[LANGFUSE_CONFIG_TIMEOUT_KEY]
    if config_values[LANGFUSE_CONFIG_DEBUG_KEY] is not None:
        os.environ["LANGFUSE_DEBUG"] = config_values[LANGFUSE_CONFIG_DEBUG_KEY]
    if config_values[LANGFUSE_CONFIG_ENVIRONMENT_KEY] is not None:
        os.environ["LANGFUSE_TRACING_ENVIRONMENT"] = config_values[
            LANGFUSE_CONFIG_ENVIRONMENT_KEY
        ]
    if config_values[LANGFUSE_CONFIG_RELEASE_KEY] is not None:
        os.environ["LANGFUSE_RELEASE"] = config_values[LANGFUSE_CONFIG_RELEASE_KEY]
    if config_values[LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY] is not None:
        os.environ["LANGFUSE_MEDIA_UPLOAD_THREAD_COUNT"] = config_values[
            LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY
        ]
    if config_values[LANGFUSE_CONFIG_SAMPLE_RATE_KEY] is not None:
        os.environ["LANGFUSE_SAMPLE_RATE"] = config_values[
            LANGFUSE_CONFIG_SAMPLE_RATE_KEY
        ]


def _configure_litellm_callback() -> None:
    """Configure LiteLLM to use Langfuse as a callback."""
    import litellm

    # LiteLLM will automatically trace LLM calls to langfuse if these environment
    # variables are set and the callback is configured
    litellm.success_callback = ["langfuse_otel"]


def _validate_langfuse_config(config_values: Dict[str, Any]) -> None:
    """Validate Langfuse configuration."""
    _validate_required_keys(
        config_values,
        [
            LANGFUSE_CONFIG_PUBLIC_KEY_KEY,
            LANGFUSE_CONFIG_PRIVATE_KEY_KEY,
            LANGFUSE_CONFIG_BASE_URL_KEY,
        ],
    )

    public_key = config_values[LANGFUSE_CONFIG_PUBLIC_KEY_KEY]
    private_key = config_values[LANGFUSE_CONFIG_PRIVATE_KEY_KEY]
    _validate_key_syntax(public_key, private_key)


def _validate_required_keys(
    config_values: Dict[str, Any], required_keys: List[str]
) -> None:
    """Validate that required keys are provided."""
    missing_keys: List[str] = []
    for key in required_keys:
        if not config_values[key]:
            missing_keys.append(key)

    if missing_keys:
        structlogger.error(
            "tracing.langfuse_config.invalid_config",
            event_info=(
                f"Langfuse config is invalid. The following keys are required and "
                f"currently not configured: {missing_keys}."
            ),
        )
        raise InvalidLangfuseConfigException(
            f"Langfuse config is invalid. The following keys are required and "
            f"currently not configured: {missing_keys}."
        )


def _validate_key_syntax(public_key: str, secret_key: str) -> None:
    """Validate that keys use the correct $syntax for environment variables."""
    if not public_key.startswith("$") or not secret_key.startswith("$"):
        structlogger.error(
            "langfuse_configuration.langfuse_config_invalid",
            event_info=(
                f"Langfuse config is invalid: {LANGFUSE_CONFIG_PUBLIC_KEY_KEY} and "
                f"{LANGFUSE_CONFIG_PRIVATE_KEY_KEY} need to be provided via the "
                "$syntax."
            ),
        )
        raise InvalidLangfuseConfigException(
            f"Langfuse config is invalid: {LANGFUSE_CONFIG_PUBLIC_KEY_KEY} and "
            f"{LANGFUSE_CONFIG_PRIVATE_KEY_KEY} need to be provided via the "
            "$syntax."
        )



import asyncio
import pprint as pretty_print
import typing
from typing import Any, Dict, Optional, Text

import rasa.core.agent
import rasa.utils.common
from rasa.shared.exceptions import RasaException
from rasa.shared.utils.cli import print_success

if typing.TYPE_CHECKING:
    from rasa.core.agent import Agent


def pprint(obj: Any) -> None:
    """Prints JSONs with indent."""
    pretty_print.pprint(obj, indent=2)


def chat(
    model_path: Optional[Text] = None,
    endpoints: Optional[Text] = None,
    agent: Optional["Agent"] = None,
) -> None:
    """Chat to the bot within a Jupyter notebook.

    Args:
        model_path: Path to a combined Rasa model.
        endpoints: Path to a yaml with the action server is custom actions are defined.
        agent: Rasa Core agent (used if no Rasa model given).
    """
    if model_path:
        agent = asyncio.run(
            rasa.core.agent.load_agent(model_path=model_path, endpoints=endpoints)
        )

    if agent is None:
        raise RasaException(
            "Either the provided model path could not load the agent "
            "or no core agent was provided."
        )

    print("Your bot is ready to talk! Type your messages here or send '/stop'.")
    while True:
        message = input()
        if message == "/stop":
            break

        responses = asyncio.run(agent.handle_text(message))
        for response in responses:
            _display_bot_response(response)


def _display_bot_response(response: Dict) -> None:
    from IPython.display import Image, display

    for response_type, value in response.items():
        if response_type == "text":
            print_success(value)

        if response_type == "image":
            image = Image(url=value)
            display(image)



# endpoints keys
ENDPOINTS_TRACING_KEY = "tracing"
ENDPOINTS_TRACING_SERVICE_NAME_KEY = "service_name"
ENDPOINTS_METRICS_KEY = "metrics"
ENDPOINTS_ROOT_CERTIFICATES_KEY = "root_certificates"
ENDPOINTS_INSECURE_KEY = "insecure"
ENDPOINTS_ENDPOINT_KEY = "endpoint"
ENDPOINTS_OTLP_BACKEND_TYPE = "otlp"

# tracing attributes
PROMPT_TOKEN_LENGTH_ATTRIBUTE_NAME = "len_prompt_tokens"
REQUEST_BODY_SIZE_IN_BYTES_ATTRIBUTE_NAME = "request_body_size_in_bytes"

# Tool output value truncation
# Truncate slot values to prevent excessive memory usage in traces
# while preserving enough context for debugging
TOOL_OUTPUT_VALUE_MAX_LENGTH = 500

# metrics constants
LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME = "llm_command_generator_cpu_usage"
LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME = "llm_command_generator_memory_usage"
LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME = (
    "llm_command_generator_prompt_token_usage"
)
LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME = (
    "llm_command_generator_llm_response_duration"
)
SINGLE_STEP_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME = (
    "single_step_llm_command_generator_cpu_usage"
)
SINGLE_STEP_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME = (
    "single_step_llm_command_generator_memory_usage"
)
SINGLE_STEP_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME = (
    "single_step_llm_command_generator_prompt_token_usage"
)
SINGLE_STEP_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME = (
    "single_step_llm_command_generator_llm_response_duration"
)
COMPACT_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME = (
    "compact_llm_command_generator_cpu_usage"
)
COMPACT_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME = (
    "compact_llm_command_generator_memory_usage"
)
COMPACT_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME = (
    "compact_llm_command_generator_prompt_token_usage"
)
COMPACT_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME = (
    "compact_llm_command_generator_llm_response_duration"
)
SEARCH_READY_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME = (
    "search_ready_llm_command_generator_cpu_usage"
)
SEARCH_READY_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME = (
    "search_ready_llm_command_generator_memory_usage"
)
SEARCH_READY_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME = (
    "search_ready_llm_command_generator_prompt_token_usage"
)
SEARCH_READY_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME = (
    "search_ready_llm_command_generator_llm_response_duration"
)
MULTI_STEP_LLM_COMMAND_GENERATOR_CPU_USAGE_METRIC_NAME = (
    "multi_step_llm_command_generator_cpu_usage"
)
MULTI_STEP_LLM_COMMAND_GENERATOR_MEMORY_USAGE_METRIC_NAME = (
    "multi_step_llm_command_generator_memory_usage"
)
MULTI_STEP_LLM_COMMAND_GENERATOR_PROMPT_TOKEN_USAGE_METRIC_NAME = (
    "multi_step_llm_command_generator_prompt_token_usage"
)
MULTI_STEP_LLM_COMMAND_GENERATOR_LLM_RESPONSE_DURATION_METRIC_NAME = (
    "multi_step_llm_command_generator_llm_response_duration"
)
ENTERPRISE_SEARCH_POLICY_CPU_USAGE_METRIC_NAME = "enterprise_search_policy_cpu_usage"
ENTERPRISE_SEARCH_POLICY_MEMORY_USAGE_METRIC_NAME = (
    "enterprise_search_policy_memory_usage"
)
ENTERPRISE_SEARCH_POLICY_PROMPT_TOKEN_USAGE_METRIC_NAME = (
    "enterprise_search_policy_prompt_token_usage"
)
ENTERPRISE_SEARCH_POLICY_LLM_RESPONSE_DURATION_METRIC_NAME = (
    "enterprise_search_policy_llm_response_duration"
)
INTENTLESS_POLICY_LLM_RESPONSE_DURATION_METRIC_NAME = (
    "intentless_policy_llm_response_duration"
)
CONTEXTUAL_RESPONSE_REPHRASER_LLM_RESPONSE_DURATION_METRIC_NAME = (
    "contextual_nlg_llm_response_duration"
)

RASA_CLIENT_REQUEST_DURATION_METRIC_NAME = "rasa_client_request_duration"
RASA_CLIENT_REQUEST_BODY_SIZE_METRIC_NAME = "rasa_client_request_body_size"

# Agent and MCP tool execution duration metrics
AGENT_EXECUTION_DURATION_METRIC_NAME = "agent_execution_duration"
MCP_TOOL_EXECUTION_DURATION_METRIC_NAME = "mcp_tool_execution_duration"

# MCP Agent LLM metrics
MCP_AGENT_LLM_CPU_USAGE_METRIC_NAME = "mcp_agent_llm_cpu_usage"
MCP_AGENT_LLM_MEMORY_USAGE_METRIC_NAME = "mcp_agent_llm_memory_usage"
MCP_AGENT_LLM_PROMPT_TOKEN_USAGE_METRIC_NAME = "mcp_agent_llm_prompt_token_usage"
MCP_AGENT_LLM_RESPONSE_DURATION_METRIC_NAME = "mcp_agent_llm_response_duration"
LLM_BASED_COMMAND_GENERATOR_CPU_MEMORY_USAGE_UNIT_NAME = "percentage"
DURATION_UNIT_NAME = "ms"

# langfuse constants
LANGFUSE_CONFIG_PUBLIC_KEY_KEY = "public_key"
LANGFUSE_CONFIG_PRIVATE_KEY_KEY = "private_key"
LANGFUSE_CONFIG_BASE_URL_KEY = "host"
LANGFUSE_CONFIG_TIMEOUT_KEY = "timeout"
LANGFUSE_CONFIG_DEBUG_KEY = "debug"
LANGFUSE_CONFIG_ENVIRONMENT_KEY = "environment"
LANGFUSE_CONFIG_RELEASE_KEY = "release"
LANGFUSE_CONFIG_MEDIA_UPLOAD_THREAD_COUNT_KEY = "media_upload_thread_count"
LANGFUSE_CONFIG_SAMPLE_RATE_KEY = "sample_rate"

# tracing types
TRACING_TYPE_LANGFUSE = "langfuse"
